package de.ozgcloud.formcycle;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.Properties;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.util.ReflectionUtils;

import de.xima.fc.interfaces.plugin.lifecycle.IPluginInitializeData;
import de.xima.fc.interfaces.workflow.value.IRecordValueDescriptorBuilder;
import de.xima.fc.interfaces.workflow.value.IStringValueDescriptor;
import de.xima.fc.interfaces.workflow.value.IValueDescriptor;
import de.xima.fc.interfaces.workflow.value.IValueDescriptorFactory;

class WorkflowElementNodePluginTest {

	@Spy
	@InjectMocks
	private WorkflowElementNodePlugin plugin;

	@Mock
	private IPluginInitializeData initializeData;
	@Mock
	private PluginPropertiesMapper propertiesMapper;

	@Nested
	class TestInitialization {

		@Test
		void shouldCallCreatePropertiesMapper() {
			plugin.initialize(initializeData);

			verify(plugin).createPropertiesMapper();
		}

		@Test
		void shouldSetInitializeData() {
			plugin.initialize(initializeData);

			assertThat(getInitializedData()).isEqualTo(initializeData);
		}

		private IPluginInitializeData getInitializedData() {
			var initializeDataField = ReflectionUtils.findField(WorkflowElementNodePlugin.class, "initializeData");
			initializeDataField.setAccessible(true);
			return (IPluginInitializeData) ReflectionUtils.getField(initializeDataField, plugin);
		}

	}

	@Nested
	@DisplayName("Success value descriptor")
	class TestSuccessValueDescriptor {

		@Mock
		private IValueDescriptorFactory descriptorFactory;
		@Mock
		private IRecordValueDescriptorBuilder descriptorBuilder;
		@Mock
		private IStringValueDescriptor stringValueDescriptor;

		@BeforeEach
		void setup() {
			when(descriptorFactory.recordBuilder()).thenReturn(descriptorBuilder);
			when(descriptorBuilder.requiredProperty(any(), any(IValueDescriptor.class))).thenReturn(descriptorBuilder);
		}

		@Test
		@DisplayName("should has vorgangnummer property")
		void shouldAddVorgangnummerProperty() {
			when(descriptorFactory.string()).thenReturn(stringValueDescriptor);

			plugin.getSuccessValueDescriptor(descriptorFactory);

			verify(descriptorBuilder).requiredProperty(ExecutionResult.VORGANGNUMMER_PROPERTY_KEY, stringValueDescriptor);
		}
	}

	@Nested
	class TestCreateOzgClient {

		@Mock
		private WorkflowElementNodeProps formProperties;

		private OzgCloudConfig ozgCloudConfig = OzgCloudConfigTestFactory.create();

		@Test
		void shouldSetOzgCloudConfig() {
			doReturn(ozgCloudConfig).when(plugin).buildOzgCloudConfig(any());

			var ozgClient = plugin.createOzgClient(formProperties);

			assertThat(getOzgCloudConfig(ozgClient)).isEqualTo(ozgCloudConfig);
		}

		@Nested
		class TestEingangsAdapterUrl {

			private static final String CUSTOM_URL = "http://custom.url";

			@Test
			void shouldReturnCustomUrl() {
				when(formProperties.getEingangsAdapterUrl()).thenReturn(CUSTOM_URL);
				when(formProperties.isUseCustomEingangsAdapterUrl()).thenReturn(true);

				var eingangsAdapterUrl = plugin.getEingangsAdapterUrl(formProperties);

				assertThat(eingangsAdapterUrl).isEqualTo(CUSTOM_URL);
			}

			@Test
			void shouldReturnDefaultUrl() {
				when(formProperties.isUseCustomEingangsAdapterUrl()).thenReturn(false);
				when(propertiesMapper.getEingangsAdapterUrl(any())).thenReturn(PropertiesTestFactory.EINGANGSADAPTER_URL);

				var eingangsAdapterUrl = plugin.getEingangsAdapterUrl(formProperties);

				assertThat(eingangsAdapterUrl).isEqualTo(PropertiesTestFactory.EINGANGSADAPTER_URL);
			}

		}

		@Nested
		class TestBuildOzgClientConfig{

			@Mock
			private Properties properties;

			@Test
			void shouldSetEingangsAdapterUrl() {
				doReturn(PropertiesTestFactory.EINGANGSADAPTER_URL).when(plugin).getEingangsAdapterUrl(any());

				var ozgClientConfig = plugin.buildOzgCloudConfig(formProperties);

				assertThat(ozgClientConfig.getEingangsAdapterUrl()).isEqualTo(PropertiesTestFactory.EINGANGSADAPTER_URL);
			}

			@Test
			void shouldSetProxyConfig() {
				var expectedProxyConfig = ProxyConfigTestFactory.create();
				when(propertiesMapper.mapProxyProperties(any())).thenReturn(expectedProxyConfig);

				var ozgClientConfig = plugin.buildOzgCloudConfig(formProperties);

				assertThat(ozgClientConfig.getProxyConfig()).isEqualTo(expectedProxyConfig);
			}

			@Test
			void shouldCallMapProxyProperties() {
				when(initializeData.getProperties()).thenReturn(properties);

				plugin.buildOzgCloudConfig(formProperties);

				verify(propertiesMapper).mapProxyProperties(properties);
			}
		}
		private OzgCloudConfig getOzgCloudConfig(OzgCloudFormDataHttpClient ozgClient) {
			var ozgCloudConfigField = ReflectionUtils.findField(OzgCloudFormDataHttpClient.class, "config");
			ozgCloudConfigField.setAccessible(true);
			return (OzgCloudConfig) ReflectionUtils.getField(ozgCloudConfigField, ozgClient);
		}
	}
}