/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch das
 * Ministerium für Energiewende, Klimaschutz, Umwelt und Natur
 * Zentrales IT-Management
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.formcycle.attachment;

import static de.ozgcloud.formcycle.attachment.FormcycleAttachmentTestFactory.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;

import de.ozgcloud.formcycle.errorhandling.NodeThrewExceptionFactory;
import de.xima.fc.entities.Attachment;
import de.xima.fc.entities.AttachmentDatei;
import de.xima.fc.exceptions.NodeThrewException;
import lombok.SneakyThrows;

class AttachmentMapperTest {

	@InjectMocks
	private AttachmentMapper mapper;

	@Mock
	private NodeThrewExceptionFactory exceptionFactory;
	@Mock
	private Attachment attachment;

	@Nested
	@DisplayName("Set fields")
	class TestSetFields {

		@Mock
		private AttachmentDatei attachmentDatei;

		@BeforeEach
		void setup() {
			when(attachmentDatei.getDaten()).thenReturn(CONTENT);
			when(attachment.getDateiName()).thenReturn(FILE_NAME);
			when(attachment.getFileEntity()).thenReturn(attachmentDatei);
		}

		@Test
		void shouldCreateAttachmentWithUuid() {
			when(attachment.getUUID()).thenReturn(UUID);

			var ozgAttachment = buildAttachments();

			assertThat(ozgAttachment).hasFieldOrPropertyWithValue(UUID_KEY, UUID);
		}

		@Test
		void shouldCreateAttachmentWithFileName() {
			var ozgAttachment = buildAttachments();

			assertThat(ozgAttachment).hasFieldOrPropertyWithValue(FILE_NAME_KEY, FILE_NAME);
		}

		@Test
		void shouldCreateAttachmentWithData() {
			var ozgAttachment = buildAttachments();

			assertThat(ozgAttachment.getContent()).hasBinaryContent(CONTENT);
		}

		@Test
		void shouldCreateAttachmentWithContentType() {
			var ozgAttachment = buildAttachments();

			assertThat(ozgAttachment.getContentType().getMimeType()).isEqualTo(FormcycleAttachmentTestFactory.CONTENT_TYPE.getMimeType());
		}

		@SneakyThrows
		private FormcycleAttachment buildAttachments() {
			return mapper.buildAttachment(attachment);
		}
	}

	@Nested
	class TestExceptions {

		@Mock
		private AttachmentDatei attachmentDatei;

		@BeforeEach
		void setup() {
			when(attachment.getFileEntity()).thenReturn(attachmentDatei);
			when(exceptionFactory.createCannotGetAttachmentException(any())).thenReturn(mock(NodeThrewException.class));
		}

		@Test
		@DisplayName("should throw when can not read attached data")
		void shouldThrowException() {
			Assertions.assertThrows(NodeThrewException.class, () -> mapper.map(attachment));

			verify(exceptionFactory).createCannotGetAttachmentException(AttachmentMapper.CANNOT_READ_ATTACHED_FILE);
		}

		@Test
		@DisplayName("should throw exception when file name is null")
		void shouldThrowExceptionName() {
			when(attachmentDatei.getDaten()).thenReturn(new byte[]{});

			Assertions.assertThrows(NodeThrewException.class, () -> mapper.map(attachment));

			verify(exceptionFactory).createCannotGetAttachmentException(AttachmentMapper.FILENAME_IS_NULL);
		}
	}
}